// -----------------------------------------------------------------------------
//    ASPiK Plugin Kernel File:  plugincore.cpp
//
/**
    \file   plugincore.cpp
    \author Will Pirkle
    \date   17-September-2018
    \brief  Implementation file for PluginCore object
    		- http://www.aspikplugins.com
    		- http://www.willpirkle.com
*/
// -----------------------------------------------------------------------------
#include "plugincore.h"
#include "plugindescription.h"

/**
\brief PluginCore constructor is launching pad for object initialization

Operations:
- initialize the plugin description (strings, codes, numbers, see initPluginDescriptors())
- setup the plugin's audio I/O channel support
- create the PluginParameter objects that represent the plugin parameters (see FX book if needed)
- create the presets
*/
PluginCore::PluginCore()
{
    // --- describe the plugin; call the helper to init the static parts you setup in plugindescription.h
    initPluginDescriptors();

    // --- default I/O combinations
	// --- for FX plugins
	if (getPluginType() == kFXPlugin)
	{
		addSupportedIOCombination({ kCFMono, kCFMono });
		addSupportedIOCombination({ kCFMono, kCFStereo });
		addSupportedIOCombination({ kCFStereo, kCFStereo });
	}
	else // --- synth plugins have no input, only output
	{
		addSupportedIOCombination({ kCFNone, kCFMono });
		addSupportedIOCombination({ kCFNone, kCFStereo });
	}

	// --- for sidechaining, we support mono and stereo inputs; auxOutputs reserved for future use
	addSupportedAuxIOCombination({ kCFMono, kCFNone });
	addSupportedAuxIOCombination({ kCFStereo, kCFNone });

	// --- create the parameters
    initPluginParameters();

    // --- create the presets
    initPluginPresets();
}

/**
\brief create all of your plugin parameters here

\return true if parameters were created, false if they already existed
*/
bool PluginCore::initPluginParameters()
{
	if (pluginParameterMap.size() > 0)
		return false;

    // --- Add your plugin parameter instantiation code bewtween these hex codes
	// **--0xDEA7--**


	// --- Declaration of Plugin Parameter Objects 
	PluginParameter* piParam = nullptr;

	// --- continuous control: Delay
	piParam = new PluginParameter(controlID::delayTime_mSec, "Delay", "mSec", controlVariableType::kDouble, 0.000000, 2000.000000, 250.000000, taper::kLinearTaper);
	piParam->setParameterSmoothing(true);
	piParam->setSmoothingTimeMsec(2000.00);
	piParam->setBoundVariable(&delayTime_mSec, boundVariableType::kDouble);
	addPluginParameter(piParam);

	// --- continuous control: Feedback
	piParam = new PluginParameter(controlID::delayFeedback_Pct, "Feedback", "%", controlVariableType::kDouble, 0.000000, 100.000000, 50.000000, taper::kLinearTaper);
	piParam->setParameterSmoothing(true);
	piParam->setSmoothingTimeMsec(20.00);
	piParam->setBoundVariable(&delayFeedback_Pct, boundVariableType::kDouble);
	addPluginParameter(piParam);

	// --- continuous control: Ratio
	piParam = new PluginParameter(controlID::delayRatio_Pct, "Ratio", "%", controlVariableType::kDouble, 1.000000, 100.000000, 50.000000, taper::kLinearTaper);
	piParam->setParameterSmoothing(true);
	piParam->setSmoothingTimeMsec(20.00);
	piParam->setBoundVariable(&delayRatio_Pct, boundVariableType::kDouble);
	addPluginParameter(piParam);

	// --- discrete control: Algorithm
	piParam = new PluginParameter(controlID::delayType, "Algorithm", "normal,pingpong", "normal");
	piParam->setBoundVariable(&delayType, boundVariableType::kInt);
	piParam->setIsDiscreteSwitch(true);
	addPluginParameter(piParam);

	// --- continuous control: Wet Level
	piParam = new PluginParameter(controlID::wetLevel_dB, "Wet Level", "dB", controlVariableType::kDouble, -60.000000, 12.000000, -3.000000, taper::kLinearTaper);
	piParam->setParameterSmoothing(true);
	piParam->setSmoothingTimeMsec(20.00);
	piParam->setBoundVariable(&wetLevel_dB, boundVariableType::kDouble);
	addPluginParameter(piParam);

	// --- continuous control: Dry Level
	piParam = new PluginParameter(controlID::dryLevel_dB, "Dry Level", "dB", controlVariableType::kDouble, -60.000000, 12.000000, -3.000000, taper::kLinearTaper);
	piParam->setParameterSmoothing(true);
	piParam->setSmoothingTimeMsec(20.00);
	piParam->setBoundVariable(&dryLevel_dB, boundVariableType::kDouble);
	addPluginParameter(piParam);

	// --- discrete control: Preset
	piParam = new PluginParameter(controlID::selectPreset, "Preset", "default,boing,gogo,whack", "default");
	piParam->setBoundVariable(&selectPreset, boundVariableType::kInt);
	piParam->setIsDiscreteSwitch(true);
	addPluginParameter(piParam);

	// --- discrete control: Dirty
	piParam = new PluginParameter(controlID::guiDirtyLED, "Dirty", "SWITCH OFF,SWITCH ON", "SWITCH OFF");
	piParam->setBoundVariable(&guiDirtyLED, boundVariableType::kInt);
	piParam->setIsDiscreteSwitch(true);
	addPluginParameter(piParam);

	// --- Aux Attributes
	AuxParameterAttribute auxAttribute;

	// --- RAFX GUI attributes
	// --- controlID::delayTime_mSec
	auxAttribute.reset(auxGUIIdentifier::guiControlData);
	auxAttribute.setUintAttribute(2147483648);
	setParamAuxAttribute(controlID::delayTime_mSec, auxAttribute);

	// --- controlID::delayFeedback_Pct
	auxAttribute.reset(auxGUIIdentifier::guiControlData);
	auxAttribute.setUintAttribute(2147483648);
	setParamAuxAttribute(controlID::delayFeedback_Pct, auxAttribute);

	// --- controlID::delayRatio_Pct
	auxAttribute.reset(auxGUIIdentifier::guiControlData);
	auxAttribute.setUintAttribute(2147483648);
	setParamAuxAttribute(controlID::delayRatio_Pct, auxAttribute);

	// --- controlID::delayType
	auxAttribute.reset(auxGUIIdentifier::guiControlData);
	auxAttribute.setUintAttribute(268435456);
	setParamAuxAttribute(controlID::delayType, auxAttribute);

	// --- controlID::wetLevel_dB
	auxAttribute.reset(auxGUIIdentifier::guiControlData);
	auxAttribute.setUintAttribute(2147483648);
	setParamAuxAttribute(controlID::wetLevel_dB, auxAttribute);

	// --- controlID::dryLevel_dB
	auxAttribute.reset(auxGUIIdentifier::guiControlData);
	auxAttribute.setUintAttribute(2147483648);
	setParamAuxAttribute(controlID::dryLevel_dB, auxAttribute);

	// --- controlID::selectPreset
	auxAttribute.reset(auxGUIIdentifier::guiControlData);
	auxAttribute.setUintAttribute(805306368);
	setParamAuxAttribute(controlID::selectPreset, auxAttribute);

	// --- controlID::guiDirtyLED
	auxAttribute.reset(auxGUIIdentifier::guiControlData);
	auxAttribute.setUintAttribute(1073741824);
	setParamAuxAttribute(controlID::guiDirtyLED, auxAttribute);


	// **--0xEDA5--**

    // --- BONUS Parameter
    // --- SCALE_GUI_SIZE
    PluginParameter* piParamBonus = new PluginParameter(SCALE_GUI_SIZE, "Scale GUI", "tiny,small,medium,normal,large,giant", "normal");
    addPluginParameter(piParamBonus);

	// --- create the super fast access array
	initPluginParameterArray();

    return true;
}

/**
\brief initialize object for a new run of audio; called just before audio streams

Operation:
- store sample rate and bit depth on audioProcDescriptor - this information is globally available to all core functions
- reset your member objects here

\param resetInfo structure of information about current audio format

\return true if operation succeeds, false otherwise
*/
bool PluginCore::reset(ResetInfo& resetInfo)
{
    // --- save for audio processing
    audioProcDescriptor.sampleRate = resetInfo.sampleRate;
    audioProcDescriptor.bitDepth = resetInfo.bitDepth;

	// --- (1) reset the component: NOTE this requires 2 steps! 
	stereoDelay.reset(resetInfo.sampleRate);

	// --- (2) create 2 second delay buffers
	stereoDelay.createDelayBuffers(resetInfo.sampleRate, 2000.0);

    // --- other reset inits
    return PluginBase::reset(resetInfo);
}

/**
\brief one-time initialize function called after object creation and before the first reset( ) call

Operation:
- saves structure for the plugin to use; you can also load WAV files or state information here
*/
bool PluginCore::initialize(PluginInfo& pluginInfo)
{
	// --- add one-time init stuff here
	//
	// --- setup presets
	//     the default preset is just the current state of the parameters at initialize time
	//	
	// --- the others are typed in manually
	presetParams[DEFAULT].algorithm = convertIntToEnum(delayType, delayAlgorithm);
	presetParams[DEFAULT].leftDelay_mSec = delayTime_mSec;
	presetParams[DEFAULT].feedback_Pct = delayFeedback_Pct;
	presetParams[DEFAULT].delayRatio_Pct = delayRatio_Pct;
	presetParams[DEFAULT].dryLevel_dB = dryLevel_dB;
	presetParams[DEFAULT].wetLevel_dB = wetLevel_dB;

	// --- short
	presetParams[BOING].algorithm = delayAlgorithm::kNormal;
	presetParams[BOING].leftDelay_mSec = 30.0;
	presetParams[BOING].feedback_Pct = 70.0;
	presetParams[BOING].delayRatio_Pct = 50.0;
	presetParams[BOING].dryLevel_dB = -3.0;
	presetParams[BOING].wetLevel_dB = -3.0;

	// --- long delay
	presetParams[GOGO].algorithm = delayAlgorithm::kPingPong;
	presetParams[GOGO].leftDelay_mSec = 1200.0;
	presetParams[GOGO].feedback_Pct = 60.0;
	presetParams[GOGO].delayRatio_Pct = 23.0;
	presetParams[GOGO].dryLevel_dB = -3.0;
	presetParams[GOGO].wetLevel_dB = -3.0;

	// --- massive FB
	presetParams[WHACK].algorithm = delayAlgorithm::kPingPong;
	presetParams[WHACK].leftDelay_mSec = 850.0;
	presetParams[WHACK].feedback_Pct = 97.0;
	presetParams[WHACK].delayRatio_Pct = 63.0;
	presetParams[WHACK].dryLevel_dB = -6.0;
	presetParams[WHACK].wetLevel_dB = -3.0;


	return true;
}


/**
\brief do anything needed prior to arrival of audio buffers

Operation:
- syncInBoundVariables when preProcessAudioBuffers is called, it is *guaranteed* that all GUI control change information
  has been applied to plugin parameters; this binds parameter changes to your underlying variables
- NOTE: postUpdatePluginParameter( ) will be called for all bound variables that are acutally updated; if you need to process
  them individually, do so in that function
- use this function to bulk-transfer the bound variable data into your plugin's member object variables

\param processInfo structure of information about *buffer* processing

\return true if operation succeeds, false otherwise
*/
bool PluginCore::preProcessAudioBuffers(ProcessBufferInfo& processInfo)
{
    // --- sync internal variables to GUI parameters; you can also do this manually if you don't
    //     want to use the auto-variable-binding
    syncInBoundVariables();

	// --- this is done in processAudioFrame() to show the difference in smoothing
	//updateParameters();

    return true;
}

/**
\brief Update C++ Object Parameters from GUI control variables

See Designing Audio Effects Plugins in C++, 2nd Edition by Will Pirkle
www.willpirkle.com

Operation:
- call getParameters on the effect object to retrieve a data structure
- set the parameters with the GUI values in the structure
- call setParameters and pass the structure back to the object

\return none
*/
	void PluginCore::updateParameters()
	{
	// --- note: ignore the preset parameter!
	AudioDelayParameters params = stereoDelay.getParameters();
	params.leftDelay_mSec = delayTime_mSec;
	params.feedback_Pct = delayFeedback_Pct;
	params.delayRatio_Pct = delayRatio_Pct;
	params.updateType = delayUpdateType::kLeftPlusRatio;
	params.dryLevel_dB = dryLevel_dB;
	params.wetLevel_dB = wetLevel_dB;

	// --- use helper
	params.algorithm = convertIntToEnum(delayType, delayAlgorithm);

	// --- set them
	stereoDelay.setParameters(params);

	}


	void PluginCore::loadPreset(uint32_t index)
	{
	if (index >= NUM_PRESETS)
		return;

	// --- grab the preset structure
	AudioDelayParameters presetParam = presetParams[index];

	// --- IF you want to sync the internal params, you can do this BUT...
	//     there is no good reason to do so using this method.
	// setPIParamValue(controlID::delayType, enumToInt(presetParam.algorithm));
	// setPIParamValue(controlID::delayTime_mSec,		delayTime_mSec);
	// etc... 

	// --- update the GUI
	HostMessageInfo hostMessageInfo;
	hostMessageInfo.hostMessage = sendGUIUpdate;

	// --- setup a GUIParameter structure
	GUIParameter guiParam[6];

	// --- algorithm
	guiParam[0].controlID = controlID::delayType;
	guiParam[0].actualValue = enumToInt(presetParam.algorithm);
	hostMessageInfo.guiUpdateData.guiParameters.push_back(guiParam[0]);

	// --- delay time
	guiParam[1].controlID = controlID::delayTime_mSec;
	guiParam[1].actualValue = presetParam.leftDelay_mSec;
	hostMessageInfo.guiUpdateData.guiParameters.push_back(guiParam[1]);

	// --- delay FB
	guiParam[2].controlID = controlID::delayFeedback_Pct;
	guiParam[2].actualValue = presetParam.feedback_Pct;
	hostMessageInfo.guiUpdateData.guiParameters.push_back(guiParam[2]);

	// --- delay ratio
	guiParam[3].controlID = controlID::delayRatio_Pct;
	guiParam[3].actualValue = presetParam.delayRatio_Pct;
	hostMessageInfo.guiUpdateData.guiParameters.push_back(guiParam[3]);

	// --- dryLevel_dB
	guiParam[4].controlID = controlID::dryLevel_dB;
	guiParam[4].actualValue = presetParam.dryLevel_dB;
	hostMessageInfo.guiUpdateData.guiParameters.push_back(guiParam[4]);

	// --- delay wetLevel_dB
	guiParam[5].controlID = controlID::wetLevel_dB;
	guiParam[5].actualValue = presetParam.wetLevel_dB;
	hostMessageInfo.guiUpdateData.guiParameters.push_back(guiParam[5]);

	// --- turn OFF the dirty GUI LED
	GUIParameter dirtyGUIParam;
	dirtyGUIParam.controlID = controlID::guiDirtyLED;
	dirtyGUIParam.actualValue = 0.0; // OFF
	hostMessageInfo.guiUpdateData.guiParameters.push_back(dirtyGUIParam);

	// --- send message to host
	pluginHostConnector->sendHostMessage(hostMessageInfo);
}


/**
\brief frame-processing method

Operation:
- decode the plugin type - for synth plugins, fill in the rendering code; for FX plugins, delete the if(synth) portion and add your processing code
- note that MIDI events are fired for each sample interval so that MIDI is tightly sunk with audio
- doSampleAccurateParameterUpdates will perform per-sample interval smoothing

\param processFrameInfo structure of information about *frame* processing

\return true if operation succeeds, false otherwise
*/
bool PluginCore::processAudioFrame(ProcessFrameInfo& processFrameInfo)
{
    // --- fire any MIDI events for this sample interval
    processFrameInfo.midiEventQueue->fireMidiEvents(processFrameInfo.currentFrame);

	// --- do per-frame updates; VST automation and parameter smoothing
	doSampleAccurateParameterUpdates();

	// --- Check is Synth Plugin and input/output existence:
	if (getPluginType() == kSynthPlugin || processFrameInfo.numAudioInChannels == 0 || processFrameInfo.numAudioOutChannels == 0)
		return false;

	// --- update GUI parameters: this is here for smoothing; you can optimize this
	//     by relocating the non-smoothed variable updates to preProcessAudioBuffers( )
	//     See the ASPiK documentation and Pirkle FX book
	updateParameters();

	// --- object does all the work on frames
	return stereoDelay.processAudioFrame(processFrameInfo.audioInputFrame,
			processFrameInfo.audioOutputFrame,
			processFrameInfo.numAudioInChannels,
			processFrameInfo.numAudioOutChannels);
}


/**
\brief do anything needed prior to arrival of audio buffers

Operation:
- updateOutBoundVariables sends metering data to the GUI meters

\param processInfo structure of information about *buffer* processing

\return true if operation succeeds, false otherwise
*/
bool PluginCore::postProcessAudioBuffers(ProcessBufferInfo& processInfo)
{
	// --- update outbound variables; currently this is meter data only, but could be extended
	//     in the future
	updateOutBoundVariables();

    return true;
}

/**
\brief update the PluginParameter's value based on GUI control, preset, or data smoothing (thread-safe)

Operation:
- update the parameter's value (with smoothing this initiates another smoothing process)
- call postUpdatePluginParameter to do any further processing

\param controlID the control ID value of the parameter being updated
\param controlValue the new control value
\param paramInfo structure of information about why this value is being udpated (e.g as a result of a preset being loaded vs. the top of a buffer process cycle)

\return true if operation succeeds, false otherwise
*/
bool PluginCore::updatePluginParameter(int32_t controlID, double controlValue, ParameterUpdateInfo& paramInfo)
{
    // --- use base class helper
    setPIParamValue(controlID, controlValue);

    // --- do any post-processing
    postUpdatePluginParameter(controlID, controlValue, paramInfo);

    return true; /// handled
}

/**
\brief update the PluginParameter's value based on *normlaized* GUI control, preset, or data smoothing (thread-safe)

Operation:
- update the parameter's value (with smoothing this initiates another smoothing process)
- call postUpdatePluginParameter to do any further processing

\param controlID the control ID value of the parameter being updated
\param normalizedValue the new control value in normalized form
\param paramInfo structure of information about why this value is being udpated (e.g as a result of a preset being loaded vs. the top of a buffer process cycle)

\return true if operation succeeds, false otherwise
*/
bool PluginCore::updatePluginParameterNormalized(int32_t controlID, double normalizedValue, ParameterUpdateInfo& paramInfo)
{
	// --- use base class helper, returns actual value
	double controlValue = setPIParamValueNormalized(controlID, normalizedValue, paramInfo.applyTaper);

	// --- do any post-processing
	postUpdatePluginParameter(controlID, controlValue, paramInfo);

	return true; /// handled
}

/**
\brief perform any operations after the plugin parameter has been updated; this is one paradigm for
	   transferring control information into vital plugin variables or member objects. If you use this
	   method you can decode the control ID and then do any cooking that is needed. NOTE: do not
	   overwrite bound variables here - this is ONLY for any extra cooking that is required to convert
	   the GUI data to meaninful coefficients or other specific modifiers.

\param controlID the control ID value of the parameter being updated
\param controlValue the new control value
\param paramInfo structure of information about why this value is being udpated (e.g as a result of a preset being loaded vs. the top of a buffer process cycle)

\return true if operation succeeds, false otherwise
*/
bool PluginCore::postUpdatePluginParameter(int32_t controlID, double controlValue, ParameterUpdateInfo& paramInfo)
{
    // --- now do any post update cooking; be careful with VST Sample Accurate automation
    //     If enabled, then make sure the cooking functions are short and efficient otherwise disable it
    //     for the Parameter involved
    /*switch(controlID)
    {
        case 0:
        {
            return true;    /// handled
        }

        default:
            return false;   /// not handled
    }*/

    return false;
}

/**
\brief has nothing to do with actual variable or updated variable (binding)

CAUTION:
- DO NOT update underlying variables here - this is only for sending GUI updates or letting you
  know that a parameter was changed; it should not change the state of your plugin.

WARNING:
- THIS IS NOT THE PREFERRED WAY TO LINK OR COMBINE CONTROLS TOGETHER. THE PROPER METHOD IS
  TO USE A CUSTOM SUB-CONTROLLER THAT IS PART OF THE GUI OBJECT AND CODE.
  SEE http://www.willpirkle.com for more information

\param controlID the control ID value of the parameter being updated
\param actualValue the new control value

\return true if operation succeeds, false otherwise
*/
bool PluginCore::guiParameterChanged(int32_t controlID, double actualValue)
{
	switch (controlID)
	{
		case controlID::selectPreset:
		{
			loadPreset(uint32_t(actualValue));

			return true; // handled
		}
		case controlID::delayFeedback_Pct:
		case controlID::delayRatio_Pct:
		case controlID::delayTime_mSec:
		case controlID::delayType:
		case controlID::dryLevel_dB:
		case controlID::wetLevel_dB:
		{
			// --- setup the message struct
			HostMessageInfo hostMessageInfo;
			hostMessageInfo.hostMessage = sendGUIUpdate;

			// --- setup a GUIParameter structure
			GUIParameter dirtyGUIParam;

			// --- turn ON the LED
			dirtyGUIParam.controlID = controlID::guiDirtyLED;
			dirtyGUIParam.actualValue = 1.0; // ON
			hostMessageInfo.guiUpdateData.guiParameters.push_back(dirtyGUIParam);

			// --- send message to host
			pluginHostConnector->sendHostMessage(hostMessageInfo);
			
			return true;
		}
	
		default:
			break;
	}

	return false; /// not handled
}

/**
\brief For Custom View and Custom Sub-Controller Operations

NOTES:
- this is for advanced users only to implement custom view and custom sub-controllers
- see the SDK for examples of use

\param messageInfo a structure containing information about the incoming message

\return true if operation succeeds, false otherwise
*/
bool PluginCore::processMessage(MessageInfo& messageInfo)
{
	// --- decode message
	switch (messageInfo.message)
	{
		// --- add customization appearance here
	case PLUGINGUI_DIDOPEN:
	{
		return false;
	}

	// --- NULL pointers so that we don't accidentally use them
	case PLUGINGUI_WILLCLOSE:
	{
		return false;
	}

	// --- update view; this will only be called if the GUI is actually open
	case PLUGINGUI_TIMERPING:
	{
		return false;
	}

	// --- register the custom view, grab the ICustomView interface
	case PLUGINGUI_REGISTER_CUSTOMVIEW:
	{

		return false;
	}

	case PLUGINGUI_REGISTER_SUBCONTROLLER:
	case PLUGINGUI_QUERY_HASUSERCUSTOM:
	case PLUGINGUI_USER_CUSTOMOPEN:
	case PLUGINGUI_USER_CUSTOMCLOSE:
	case PLUGINGUI_EXTERNAL_SET_NORMVALUE:
	case PLUGINGUI_EXTERNAL_SET_ACTUALVALUE:
	{

		return false;
	}

	default:
		break;
	}

	return false; /// not handled
}


/**
\brief process a MIDI event

NOTES:
- MIDI events are 100% sample accurate; this function will be called repeatedly for every MIDI message
- see the SDK for examples of use

\param event a structure containing the MIDI event data

\return true if operation succeeds, false otherwise
*/
bool PluginCore::processMIDIEvent(midiEvent& event)
{
	return true;
}

/**
\brief (for future use)

NOTES:
- MIDI events are 100% sample accurate; this function will be called repeatedly for every MIDI message
- see the SDK for examples of use

\param vectorJoysickData a structure containing joystick data

\return true if operation succeeds, false otherwise
*/
bool PluginCore::setVectorJoystickParameters(const VectorJoystickData& vectorJoysickData)
{
	return true;
}

/**
\brief use this method to add new presets to the list

NOTES:
- see the SDK for examples of use
- for non RackAFX users that have large paramter counts, there is a secret GUI control you
  can enable to write C++ code into text files, one per preset. See the SDK or http://www.willpirkle.com for details

\return true if operation succeeds, false otherwise
*/
bool PluginCore::initPluginPresets()
{
	// **--0xFF7A--**

	// --- Plugin Presets 
	int index = 0;
	PresetInfo* preset = nullptr;

	// --- Preset: Factory Preset
	preset = new PresetInfo(index++, "Factory Preset");
	initPresetParameters(preset->presetParameters);
	setPresetParameter(preset->presetParameters, controlID::delayTime_mSec, 250.000000);
	setPresetParameter(preset->presetParameters, controlID::delayFeedback_Pct, 50.000000);
	setPresetParameter(preset->presetParameters, controlID::delayRatio_Pct, 50.000000);
	setPresetParameter(preset->presetParameters, controlID::delayType, -0.000000);
	setPresetParameter(preset->presetParameters, controlID::wetLevel_dB, -3.000000);
	setPresetParameter(preset->presetParameters, controlID::dryLevel_dB, -3.000000);
	setPresetParameter(preset->presetParameters, controlID::selectPreset, -0.000000);
	setPresetParameter(preset->presetParameters, controlID::guiDirtyLED, -0.000000);
	addPreset(preset);

	// --- Preset: goofgoooa
	preset = new PresetInfo(index++, "goofgoooa");
	initPresetParameters(preset->presetParameters);
	setPresetParameter(preset->presetParameters, controlID::delayTime_mSec, 1200.000000);
	setPresetParameter(preset->presetParameters, controlID::delayFeedback_Pct, 60.000004);
	setPresetParameter(preset->presetParameters, controlID::delayRatio_Pct, 23.000000);
	setPresetParameter(preset->presetParameters, controlID::delayType, 1.000000);
	setPresetParameter(preset->presetParameters, controlID::wetLevel_dB, -3.000000);
	setPresetParameter(preset->presetParameters, controlID::dryLevel_dB, -3.000000);
	setPresetParameter(preset->presetParameters, controlID::selectPreset, 2.000000);
	setPresetParameter(preset->presetParameters, controlID::guiDirtyLED, 0.000000);
	addPreset(preset);


	// **--0xA7FF--**

    return true;
}

/**
\brief setup the plugin description strings, flags and codes; this is ordinarily done through the ASPiKreator or CMake

\return true if operation succeeds, false otherwise
*/
bool PluginCore::initPluginDescriptors()
{
    pluginDescriptor.pluginName = PluginCore::getPluginName();
    pluginDescriptor.shortPluginName = PluginCore::getShortPluginName();
    pluginDescriptor.vendorName = PluginCore::getVendorName();
    pluginDescriptor.pluginTypeCode = PluginCore::getPluginType();

	// --- describe the plugin attributes; set according to your needs
	pluginDescriptor.hasSidechain = kWantSidechain;
	pluginDescriptor.latencyInSamples = kLatencyInSamples;
	pluginDescriptor.tailTimeInMSec = kTailTimeMsec;
	pluginDescriptor.infiniteTailVST3 = kVSTInfiniteTail;

    // --- AAX
    apiSpecificInfo.aaxManufacturerID = kManufacturerID;
    apiSpecificInfo.aaxProductID = kAAXProductID;
    apiSpecificInfo.aaxBundleID = kAAXBundleID;  /* MacOS only: this MUST match the bundle identifier in your info.plist file */
    apiSpecificInfo.aaxEffectID = "aaxDeveloper.";
    apiSpecificInfo.aaxEffectID.append(PluginCore::getPluginName());
    apiSpecificInfo.aaxPluginCategoryCode = kAAXCategory;

    // --- AU
    apiSpecificInfo.auBundleID = kAUBundleID;
	apiSpecificInfo.auBundleName = kAUBundleName;   /* MacOS only: this MUST match the bundle identifier in your info.plist file */

    // --- VST3
    apiSpecificInfo.vst3FUID = PluginCore::getVSTFUID(); // OLE string format
    apiSpecificInfo.vst3BundleID = kVST3BundleID;/* MacOS only: this MUST match the bundle identifier in your info.plist file */
	apiSpecificInfo.enableVST3SampleAccurateAutomation = kVSTSAA;
	apiSpecificInfo.vst3SampleAccurateGranularity = kVST3SAAGranularity;

    // --- AU and AAX
    apiSpecificInfo.fourCharCode = PluginCore::getFourCharCode();

    return true;
}

// --- static functions required for VST3/AU only --------------------------------------------- //
const char* PluginCore::getPluginBundleName() { return kAUBundleName; }
const char* PluginCore::getPluginName(){ return kPluginName; }
const char* PluginCore::getShortPluginName(){ return kShortPluginName; }
const char* PluginCore::getVendorName(){ return kVendorName; }
const char* PluginCore::getVendorURL(){ return kVendorURL; }
const char* PluginCore::getVendorEmail(){ return kVendorEmail; }
const char* PluginCore::getAUCocoaViewFactoryName(){ return AU_COCOA_VIEWFACTORY_STRING; }
pluginType PluginCore::getPluginType(){ return kPluginType; }
const char* PluginCore::getVSTFUID(){ return kVSTFUID; }
int32_t PluginCore::getFourCharCode(){ return kFourCharCode; }
